@@ -14,6 +14,11 @@
 
 #include <jni.h>
 #include <stdarg.h>
+#ifdef _MSC_VER
+#include <Winsock2.h>
+#else
+#include <arpa/inet.h>
+#endif
 
 #include "java/lang/Object.h"
 #include "java/lang/Class.h"
@@ -95,6 +100,170 @@ PyObject *findClass(PyObject *self, PyObject *args)
     Py_RETURN_NONE;
 }
 
+static const char interface_bytes[] = {
+    '\xca', '\xfe', '\xba', '\xbe',   // magic number: 0xcafebabe
+    '\x00', '\x00', '\x00', '\x32',   // version 50.0
+    '\x00', '\x07',                   // constant pool max index: 6
+    '\x07', '\x00', '\x04',           // 1: class name at 4
+    '\x07', '\x00', '\x05',           // 2: class name at 5
+    '\x07', '\x00', '\x06',           // 3: class name at 6
+    '\x01', '\x00', '\x00',           // 4: empty string
+    '\x01', '\x00', '\x10',           // 5: 16-byte string: java/lang/Object
+    'j', 'a', 'v', 'a', '/', 'l', 'a', 'n', 'g', '/',
+    'O', 'b', 'j', 'e', 'c', 't',
+    '\x01', '\x00', '\x00',           // 6: empty string
+    '\x06', '\x01',                   // public abstract interface
+    '\x00', '\x01',                   // this class at 1
+    '\x00', '\x02',                   // superclass at 2
+    '\x00', '\x01',                   // 1 interface
+    '\x00', '\x03',                   // interface at 3
+    '\x00', '\x00',                   // 0 fields
+    '\x00', '\x00',                   // 0 methods
+    '\x00', '\x00'                    // 0 attributes
+};
+
+/* make an empty interface that extends an interface */
+PyObject *makeInterface(PyObject *self, PyObject *args)
+{
+    char *name, *extName;
+    int name_len, extName_len;
+
+    if (!PyArg_ParseTuple(args, "s#s#",
+                          &name, &name_len, &extName, &extName_len))
+        return NULL;
+    
+    JNIEnv *vm_env = env->get_vm_env();
+    jclass _ucl = (jclass) vm_env->FindClass("java/net/URLClassLoader");
+    jmethodID mid = vm_env->GetStaticMethodID(_ucl, "getSystemClassLoader",
+                                              "()Ljava/lang/ClassLoader;");
+    jobject classLoader = vm_env->CallStaticObjectMethod(_ucl, mid);
+    const int bytes_len = sizeof(interface_bytes);
+    const int len = bytes_len + name_len + extName_len;
+    char *buf = (char *) malloc(len);
+
+    if (buf == NULL)
+        return PyErr_NoMemory();
+
+    int name_pos = 22;
+    int extName_pos = 44;
+    jclass cls;
+
+    memcpy(buf, interface_bytes, name_pos);
+    memcpy(buf + name_pos + name_len, interface_bytes + name_pos,
+           extName_pos - name_pos);
+    memcpy(buf + extName_pos + name_len + extName_len,
+           interface_bytes + extName_pos, bytes_len - extName_pos);
+    extName_pos += name_len;
+
+    *((unsigned short *) (buf + name_pos - 2)) = htons(name_len);
+    memcpy(buf + name_pos, name, name_len);
+
+    *((unsigned short *) (buf + extName_pos - 2)) = htons(extName_len);
+    memcpy(buf + extName_pos, extName, extName_len);
+
+    cls = vm_env->DefineClass(name, classLoader, (const jbyte *) buf, len);
+    free(buf);
+
+    if (cls)
+        return t_Class::wrap_Object(Class(cls));
+
+    return PyErr_SetJavaError();
+}
+
+static const char class_bytes[] = {
+    '\xca', '\xfe', '\xba', '\xbe',          // magic number: 0xcafebabe
+    '\x00', '\x00', '\x00', '\x32',          // version 50.0
+    '\x00', '\x0c',                          // constant pool max index: 11
+    '\x0a', '\x00', '\x03', '\x00', '\x08',  // 1: method for class 3 at 8
+    '\x07', '\x00', '\x09',                  // 2: class name at 9
+    '\x07', '\x00', '\x0a',                  // 3: class name at 10
+    '\x07', '\x00', '\x0b',                  // 4: class name at 11
+    '\x01', '\x00', '\x06',                  // 5: 6-byte string: <init>
+    '<', 'i', 'n', 'i', 't', '>',
+    '\x01', '\x00', '\x03', '(', ')', 'V',   // 6: 3-byte string: ()V
+    '\x01', '\x00', '\x04',                  // 7: 4-byte string: Code
+    'C', 'o', 'd', 'e',
+    '\x0c', '\x00', '\x05', '\x00', '\x06',  // 8: name at 5, signature at 6
+    '\x01', '\x00', '\x00',                  // 9: empty string
+    '\x01', '\x00', '\x00',                  // 10: empty string
+    '\x01', '\x00', '\x00',                  // 11: empty string
+    '\x00', '\x21',                          // super public
+    '\x00', '\x02',                          // this class at 2
+    '\x00', '\x03',                          // superclass at 3
+    '\x00', '\x01',                          // 1 interface
+    '\x00', '\x04',                          // interface at 4
+    '\x00', '\x00',                          // 0 fields
+    '\x00', '\x01',                          // 1 method
+    '\x00', '\x01', '\x00', '\x05',          // public, name at 5
+    '\x00', '\x06', '\x00', '\x01',          // signature at 6, 1 attribute
+    '\x00', '\x07',                          // attribute name at 7: Code
+    '\x00', '\x00', '\x00', '\x11',          // 17 bytes past 6 attribute bytes
+    '\x00', '\x01',                          // max stack: 1
+    '\x00', '\x01',                          // max locals: 1
+    '\x00', '\x00', '\x00', '\x05',          // code length: 5
+    '\x2a', '\xb7', '\x00', '\x01', '\xb1',  // actual code bytes
+    '\x00', '\x00',                          // 0 method exceptions
+    '\x00', '\x00',                          // 0 method attributes
+    '\x00', '\x00',                          // 0 attributes
+};
+
+/* make an empty class that extends a class and implements an interface */
+PyObject *makeClass(PyObject *self, PyObject *args)
+{
+    char *name, *extName, *implName;
+    int name_len, extName_len, implName_len;
+
+    if (!PyArg_ParseTuple(args, "s#s#s#",
+                          &name, &name_len, &extName, &extName_len,
+                          &implName, &implName_len))
+        return NULL;
+    
+    JNIEnv *vm_env = env->get_vm_env();
+    jclass _ucl = (jclass) vm_env->FindClass("java/net/URLClassLoader");
+    jmethodID mid = vm_env->GetStaticMethodID(_ucl, "getSystemClassLoader",
+                                              "()Ljava/lang/ClassLoader;");
+    jobject classLoader = vm_env->CallStaticObjectMethod(_ucl, mid);
+    const int bytes_len = sizeof(class_bytes);
+    const int len = bytes_len + name_len + extName_len + implName_len;
+    char *buf = (char *) malloc(len);
+
+    if (buf == NULL)
+        return PyErr_NoMemory();
+
+    int name_pos = 54;
+    int extName_pos = 57;
+    int implName_pos = 60;
+    jclass cls;
+
+    memcpy(buf, class_bytes, name_pos);
+    memcpy(buf + name_pos + name_len, class_bytes + name_pos,
+           extName_pos - name_pos);
+    memcpy(buf + extName_pos + name_len + extName_len,
+           class_bytes + extName_pos, bytes_len - extName_pos);
+    memcpy(buf + implName_pos + name_len + extName_len + implName_len,
+           class_bytes + implName_pos, bytes_len - implName_pos);
+
+    extName_pos += name_len;
+    implName_pos += name_len + extName_len;
+
+    *((unsigned short *) (buf + name_pos - 2)) = htons(name_len);
+    memcpy(buf + name_pos, name, name_len);
+
+    *((unsigned short *) (buf + extName_pos - 2)) = htons(extName_len);
+    memcpy(buf + extName_pos, extName, extName_len);
+
+    *((unsigned short *) (buf + implName_pos - 2)) = htons(implName_len);
+    memcpy(buf + implName_pos, implName, implName_len);
+
+    cls = vm_env->DefineClass(name, classLoader, (const jbyte *) buf, len);
+    free(buf);
+
+    if (cls)
+        return t_Class::wrap_Object(Class(cls));
+
+    return PyErr_SetJavaError();
+}
+
 static boxfn get_boxfn(PyTypeObject *type)
 {
     static PyObject *boxfn_ = PyString_FromString("boxfn_");
@@ -165,9 +334,6 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
     unsigned int typeCount = strlen(types);
     va_list list, check;
 
-    if (count > typeCount)
-        return -1;
-
     va_start(list, types);
     va_start(check, types);
 #endif
@@ -186,17 +352,27 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
         return -1;
     }
 
+    bool last = false;      /* true if last expected parameter */
+    bool varargs = false;   /* true if in varargs mode */
+    bool empty = false;     /* true if in varargs mode and no params passed */
+    int array = 0;          /* > 0 if expecting an array, its nesting level */
     unsigned int pos = 0;
-    int array = 0;
 
     for (unsigned int a = 0; a < count; a++, pos++) {
         PyObject *arg = args[a];
+        char tc = types[pos];
 
-        switch (types[pos]) {
+        last = last || types[pos + 1] == '\0';
+
+        if (array > 1 && tc != '[')
+          tc = 'o';
+
+        switch (tc) {
           case '[':
           {
-              if (++array > 1)
-                  return -1;
+              if (++array > 1 &&
+                  !PyObject_TypeCheck(arg, PY_TYPE(JArrayObject)))
+                return -1;
 
               a -= 1;
               break;
@@ -208,7 +384,7 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
           {
               jclass cls = NULL;
 
-              switch (types[pos]) {
+              switch (tc) {
                 case 'j':
                   cls = (jclass) va_arg(list, Class *)->this$;
                   break;
@@ -216,7 +392,7 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
                 case 'K':
                   try {
                       getclassfn initializeClass = va_arg(list, getclassfn);
-                      cls = (*initializeClass)();
+                      cls = env->getClass(initializeClass);
                   } catch (int e) {
                       switch (e) {
                         case _EXC_PYTHON:
@@ -238,7 +414,7 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
                * case because of earlier recursion avoidance (JObject(cls)).
                */
               if (!Class::class$)
-                  Class::initializeClass();
+                  env->getClass(Class::initializeClass);
 
               if (array)
               {
@@ -274,6 +450,30 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
                       else
                           break;
                   }
+
+                  if (last)
+                  {
+                      int ok = 0;
+
+                      if (arg == Py_None)
+                          ok = 1;
+                      else if (PyObject_TypeCheck(arg, &PY_TYPE(Object)) &&
+                               vm_env->IsInstanceOf(((t_Object *) arg)->object.this$, cls))
+                          ok = 1;
+                      else if (PyObject_TypeCheck(arg, &PY_TYPE(FinalizerProxy)))
+                      {
+                          PyObject *o = ((t_fp *) arg)->object;
+
+                          if (PyObject_TypeCheck(o, &PY_TYPE(Object)) &&
+                              vm_env->IsInstanceOf(((t_Object *) o)->object.this$, cls))
+                              ok = 1;
+                      }
+                      if (ok)
+                      {
+                          varargs = true;
+                          break;
+                      }
+                  }
               }
               else if (PyObject_TypeCheck(arg, &PY_TYPE(Object)) &&
                        vm_env->IsInstanceOf(((t_Object *) arg)->object.this$, cls))
@@ -313,6 +513,12 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
                       else
                           break;
                   }
+
+                  if (last && (arg == Py_True || arg == Py_False))
+                  {
+                      varargs = true;
+                      break;
+                  }
               }
               else if (arg == Py_True || arg == Py_False)
                   break;
@@ -328,6 +534,16 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
                       break;
                   if (PyObject_TypeCheck(arg, PY_TYPE(JArrayByte)))
                       break;
+
+                  if (last)
+                  {
+                      if ((PyString_Check(arg) && (PyString_Size(arg) == 1)) ||
+                          PyInt_CheckExact(arg))
+                      {
+                          varargs = true;
+                          break;
+                      }
+                  }
               }
               else if (PyString_Check(arg) && (PyString_Size(arg) == 1))
                   break;
@@ -345,6 +561,15 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
                       break;
                   if (PyObject_TypeCheck(arg, PY_TYPE(JArrayChar)))
                       break;
+
+                  if (last)
+                  {
+                      if (PyUnicode_Check(arg) && (PyUnicode_GET_SIZE(arg) == 1))
+                      {
+                          varargs = true;
+                          break;
+                      }
+                  }
               }
               else if (PyUnicode_Check(arg) && PyUnicode_GET_SIZE(arg) == 1)
                   break;
@@ -375,6 +600,12 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
                       else
                           break;
                   }
+
+                  if (last && PyInt_CheckExact(arg))
+                  {
+                      varargs = true;
+                      break;
+                  }
               }
               else if (PyInt_CheckExact(arg))
                   break;
@@ -406,6 +637,12 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
                       else
                           break;
                   }
+
+                  if (last && PyInt_CheckExact(arg))
+                  {
+                      varargs = true;
+                      break;
+                  }
               }
               else if (PyInt_CheckExact(arg))
                   break;
@@ -437,6 +674,12 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
                       else
                           break;
                   }
+
+                  if (last && PyFloat_CheckExact(arg))
+                  {
+                      varargs = true;
+                      break;
+                  }
               }
               else if (PyFloat_CheckExact(arg))
                   break;
@@ -468,6 +711,12 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
                       else
                           break;
                   }
+
+                  if (last && PyFloat_CheckExact(arg))
+                  {
+                      varargs = true;
+                      break;
+                  }
               }
               else if (PyFloat_CheckExact(arg))
                   break;
@@ -499,6 +748,12 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
                       else
                           break;
                   }
+
+                  if (last && PyLong_CheckExact(arg))
+                  {
+                      varargs = true;
+                      break;
+                  }
               }
               else if (PyLong_CheckExact(arg))
                   break;
@@ -532,6 +787,13 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
                       }
                       else
                           break;
+                  } 
+
+                  if (last && (arg == Py_None ||
+                               PyString_Check(arg) || PyUnicode_Check(arg)))
+                  {
+                      varargs = true;
+                      break;
                   }
               }
               else if (arg == Py_None ||
@@ -585,25 +847,57 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
             return -1;
         }
 
-        if (types[pos] != '[')
+        if (tc != '[')
             array = 0;
+
+        if (varargs)
+        {
+            pos = typeCount;
+            break;
+        }
     }
 
     if (array)
         return -1;
 
-    pos = 0;
+    if (pos == typeCount - 2 && types[pos] == '[' && types[pos + 1] != '[')
+    {
+        varargs = true;
+        empty = true;
+        pos = typeCount;
+    }
 
-    for (unsigned int a = 0; a < count; a++, pos++) {
-        PyObject *arg = args[a];
-        
-        switch (types[pos]) {
+    if (pos != typeCount)
+      return -1;
+
+    pos = 0;
+    last = false;
+
+    for (unsigned int a = 0; a <= count; a++, pos++) {
+        char tc = types[pos];
+        PyObject *arg;
+
+        if (a == count)
+        {
+            if (empty)  /* empty varargs */
+                arg = NULL;
+            else
+                break;
+        }
+        else
+            arg = args[a];
+
+        last = last || types[pos + 1] == '\0';
+
+        if (array > 1 && tc != '[')
+            tc = 'o';
+
+        switch (tc) {
           case '[':
           {
-              if (++array > 1)
-                  return -1;
-
+              array += 1;
               a -= 1;
+
               break;
           }
 
@@ -613,14 +907,14 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
           {
               jclass cls = NULL;
 
-              switch (types[pos]) {
+              switch (tc) {
                 case 'j':
                   cls = (jclass) va_arg(check, Class *)->this$;
                   break;
                 case 'k':
                 case 'K':
                   getclassfn initializeClass = va_arg(check, getclassfn);
-                  cls = (*initializeClass)();
+                  cls = env->getClass(initializeClass);
                   break;
               }
 
@@ -629,7 +923,7 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
                   JArray<jobject> *array = va_arg(list, JArray<jobject> *);
 
 #ifdef _java_generics
-                  if (types[pos] == 'K')
+                  if (tc == 'K')
                   {
                       PyTypeObject ***tp = va_arg(list, PyTypeObject ***);
 
@@ -639,6 +933,8 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
 #endif
                   if (arg == Py_None)
                       *array = JArray<jobject>((jobject) NULL);
+                  else if (last && varargs)
+                      *array = fromPySequence(cls, args + a, count - a);
                   else if (PyObject_TypeCheck(arg, PY_TYPE(JArrayObject)))
                       *array = ((t_JArray<jobject> *) arg)->array;
                   else 
@@ -655,7 +951,7 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
                       arg = ((t_fp *) arg)->object;
 
 #ifdef _java_generics
-                  if (types[pos] == 'K')
+                  if (tc == 'K')
                   {
                       PyTypeObject ***tp = va_arg(list, PyTypeObject ***);
                       PyTypeObject **(*parameters_)(void *) = 
@@ -683,6 +979,8 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
 
                   if (arg == Py_None)
                       *array = JArray<jboolean>((jobject) NULL);
+                  else if (last && varargs)
+                      *array = JArray<jboolean>(args + a, count - a);
                   else if (PyObject_TypeCheck(arg, PY_TYPE(JArrayBool)))
                       *array = ((t_JArray<jboolean> *) arg)->array;
                   else
@@ -707,6 +1005,8 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
 
                   if (arg == Py_None)
                       *array = JArray<jbyte>((jobject) NULL);
+                  else if (last && varargs)
+                      *array = JArray<jbyte>(args + a, count - a);
                   else if (PyObject_TypeCheck(arg, PY_TYPE(JArrayByte)))
                       *array = ((t_JArray<jbyte> *) arg)->array;
                   else 
@@ -736,6 +1036,8 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
 
                   if (arg == Py_None)
                       *array = JArray<jchar>((jobject) NULL);
+                  else if (last && varargs)
+                      *array = JArray<jchar>(args + a, count - a);
                   else if (PyObject_TypeCheck(arg, PY_TYPE(JArrayChar)))
                       *array = ((t_JArray<jchar> *) arg)->array;
                   else 
@@ -760,6 +1062,8 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
 
                   if (arg == Py_None)
                       *array = JArray<jint>((jobject) NULL);
+                  else if (last && varargs)
+                      *array = JArray<jint>(args + a, count - a);
                   else if (PyObject_TypeCheck(arg, PY_TYPE(JArrayInt)))
                       *array = ((t_JArray<jint> *) arg)->array;
                   else 
@@ -784,6 +1088,8 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
 
                   if (arg == Py_None)
                       *array = JArray<jshort>((jobject) NULL);
+                  else if (last && varargs)
+                      *array = JArray<jshort>(args + a, count - a);
                   else if (PyObject_TypeCheck(arg, PY_TYPE(JArrayShort)))
                       *array = ((t_JArray<jshort> *) arg)->array;
                   else 
@@ -808,6 +1114,8 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
 
                   if (arg == Py_None)
                       *array = JArray<jdouble>((jobject) NULL);
+                  else if (last && varargs)
+                      *array = JArray<jdouble>(args + a, count - a);
                   else if (PyObject_TypeCheck(arg, PY_TYPE(JArrayDouble)))
                       *array = ((t_JArray<jdouble> *) arg)->array;
                   else 
@@ -832,6 +1140,8 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
 
                   if (arg == Py_None)
                       *array = JArray<jfloat>((jobject) NULL);
+                  else if (last && varargs)
+                      *array = JArray<jfloat>(args + a, count - a);
                   else if (PyObject_TypeCheck(arg, PY_TYPE(JArrayFloat)))
                       *array = ((t_JArray<jfloat> *) arg)->array;
                   else 
@@ -856,6 +1166,8 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
 
                   if (arg == Py_None)
                       *array = JArray<jlong>((jobject) NULL);
+                  else if (last && varargs)
+                      *array = JArray<jlong>(args + a, count - a);
                   else if (PyObject_TypeCheck(arg, PY_TYPE(JArrayLong)))
                       *array = ((t_JArray<jlong> *) arg)->array;
                   else 
@@ -880,9 +1192,13 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
 
                   if (arg == Py_None)
                       *array = JArray<jstring>((jobject) NULL);
+                  else if (last && varargs)
+                      *array = fromPySequence(
+                          env->getClass(String::initializeClass),
+                          args + a, count - a);
                   else if (PyObject_TypeCheck(arg, PY_TYPE(JArrayString)))
                       *array = ((t_JArray<jstring> *) arg)->array;
-                  else 
+                  else
                       *array = JArray<jstring>(arg);
 
                   if (PyErr_Occurred())
@@ -913,6 +1229,10 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
 
                   if (arg == Py_None)
                       *array = JArray<Object>((jobject) NULL);
+                  else if (last && varargs)
+                      *array = JArray<Object>(
+                          env->getClass(Object::initializeClass),
+                          args + a, count - a);
                   else 
                       *array = JArray<Object>(arg);
 
@@ -923,7 +1243,7 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
               {
                   Object *obj = va_arg(list, Object *);
 
-                  if (types[pos] == 'O')
+                  if (tc == 'O')
                   {
                       PyTypeObject *type = va_arg(check, PyTypeObject *);
 
@@ -964,8 +1284,14 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
             return -1;
         }
 
-        if (types[pos] != '[')
+        if (tc != '[')
             array = 0;
+
+        if (last && varargs)
+        {
+            pos = typeCount;
+            break;
+        }
     }
 
     if (pos == typeCount)
@@ -1227,6 +1553,69 @@ PyObject *get_extension_nextElement(PyObject *self)
     return PyObject_CallMethod(self, "nextElement", "");
 }
 
+static bool setArrayObj(jobjectArray array, int index, PyObject *obj)
+{
+    bool deleteLocal = false;
+    jobject jobj;
+
+    if (obj == Py_None)
+      jobj = NULL;
+    else if (PyString_Check(obj) || PyUnicode_Check(obj))
+    {
+        jobj = env->fromPyString(obj);
+        deleteLocal = true;
+    }
+    else if (PyObject_TypeCheck(obj, &PY_TYPE(JObject)))
+        jobj = ((t_JObject *) obj)->object.this$;
+    else if (PyObject_TypeCheck(obj, &PY_TYPE(FinalizerProxy)))
+        jobj = ((t_JObject *) ((t_fp *) obj)->object)->object.this$;
+    else if (obj == Py_True || obj == Py_False)
+    {
+        jobj = env->boxBoolean(obj == Py_True);
+        deleteLocal = true;
+    }
+    else if (PyFloat_Check(obj))
+    {
+        jobj = env->boxDouble(PyFloat_AS_DOUBLE(obj));
+        deleteLocal = true;
+    }
+    else if (PyInt_Check(obj))
+    {
+        jobj = env->boxInteger(PyInt_AS_LONG(obj));
+        deleteLocal = true;
+    }
+    else if (PyLong_Check(obj))
+    {
+        jobj = env->boxLong(PyLong_AsLongLong(obj));
+        deleteLocal = true;
+    }
+    else
+    {
+        PyErr_SetObject(PyExc_TypeError, obj);
+        Py_DECREF(obj);
+        return false;
+    }
+
+    try {
+        env->setObjectArrayElement(array, index, jobj);
+        if (deleteLocal)
+            env->get_vm_env()->DeleteLocalRef(jobj);
+        Py_DECREF(obj);
+    } catch (int e) {
+        Py_DECREF(obj);
+        switch (e) {
+          case _EXC_JAVA:
+            PyErr_SetJavaError();
+            return false;
+          default:
+            throw;
+        }
+    }
+
+    return true;
+}
+
+
 jobjectArray fromPySequence(jclass cls, PyObject *sequence)
 {
     if (sequence == Py_None)
@@ -1255,50 +1644,48 @@ jobjectArray fromPySequence(jclass cls, PyObject *sequence)
         }
     }
 
-    JNIEnv *vm_env = env->get_vm_env();
-
     for (int i = 0; i < length; i++) {
         PyObject *obj = PySequence_GetItem(sequence, i);
-        int fromString = 0;
-        jobject jobj;
 
         if (!obj)
             break;
-        else if (obj == Py_None)
-            jobj = NULL;
-        else if (PyString_Check(obj) || PyUnicode_Check(obj))
-        {
-            jobj = env->fromPyString(obj);
-            fromString = 1;
-        }
-        else if (PyObject_TypeCheck(obj, &PY_TYPE(JObject)))
-            jobj = ((t_JObject *) obj)->object.this$;
-        else if (PyObject_TypeCheck(obj, &PY_TYPE(FinalizerProxy)))
-            jobj = ((t_JObject *) ((t_fp *) obj)->object)->object.this$;
-        else /* todo: add auto-boxing of primitive types */
-        {
-            PyErr_SetObject(PyExc_TypeError, obj);
-            Py_DECREF(obj);
+
+        if (!setArrayObj(array, i, obj))
             return NULL;
-        }
+    }
 
-        Py_DECREF(obj);
+    return array;
+}
 
-        try {
-            env->setObjectArrayElement(array, i, jobj);
-            if (fromString)
-                vm_env->DeleteLocalRef(jobj);
-        } catch (int e) {
-            switch (e) {
-              case _EXC_JAVA:
-                PyErr_SetJavaError();
-                return NULL;
-              default:
-                throw;
-            }
+jobjectArray fromPySequence(jclass cls, PyObject **args, int length)
+{
+    jobjectArray array;
+
+    try {
+        array = env->newObjectArray(cls, length);
+    } catch (int e) {
+        switch (e) {
+          case _EXC_PYTHON:
+            return NULL;
+          case _EXC_JAVA:
+            PyErr_SetJavaError();
+            return NULL;
+          default:
+            throw;
         }
     }
 
+    for (int i = 0; i < length; i++) {
+        PyObject *obj = args[i];
+
+        if (!obj)
+            break;
+
+        Py_INCREF(obj);
+        if (!setArrayObj(array, i, obj))
+            return NULL;
+    }
+
     return array;
 }
 
